<?php

namespace Drupal\lite_youtube\Plugin\Field\FieldFormatter;

use Drupal\Component\Plugin\Exception\PluginNotFoundException;
use Drupal\Core\Field\FieldDefinitionInterface;
use Drupal\Core\Field\FieldItemListInterface;
use Drupal\Core\Field\FormatterBase;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Logger\LoggerChannelFactoryInterface;
use Drupal\media\Plugin\media\Source\OEmbedInterface;
use Drupal\youtube\Plugin\Field\FieldType\YouTubeItem;
use Psr\Log\LoggerInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Plugin implementation of the 'lite_youtube' formatter.
 *
 * This formatter processes YouTube video fields using a lightweight
 * Implementation that improves performance by loading videos deferred.
 *
 * Key features:
 * - Lazy loading of YouTube videos for better performance
 * - Support for multiple field types (youtube, link, string, string_long)
 * - Configurable thumbnail quality settings
 * - Nocookie option for enhanced privacy
 * - YouTube Shorts support
 * - Automatic loading with Intersection Observer
 *
 * @FieldFormatter(
 *   id = "lite_youtube",
 *   label = @Translation("Lite YouTube"),
 *   field_types = {
 *     "youtube",
 *     "link",
 *     "string",
 *     "string_long",
 *   },
 * )
 */
class LiteYoutubeFormatter extends FormatterBase {

  /**
   * The logger service.
   */
  protected LoggerInterface $logger;

  /**
   * {@inheritdoc}
   */
  public function __construct($plugin_id, $plugin_definition, FieldDefinitionInterface $field_definition, array $settings, $label, $view_mode, array $third_party_settings, LoggerChannelFactoryInterface $logger_factory) {
    parent::__construct($plugin_id, $plugin_definition, $field_definition, $settings, $label, $view_mode, $third_party_settings);
    $this->logger = $logger_factory->get('media');
  }

  /**
   * {@inheritdoc}
   *
   * @return static
   *   Returns an instance of this plugin.
   */
  public static function create(ContainerInterface $container, array $configuration, $plugin_id, $plugin_definition) {
    return new static(
          $plugin_id,
          $plugin_definition,
          $configuration['field_definition'],
          $configuration['settings'],
          $configuration['label'],
          $configuration['view_mode'],
          $configuration['third_party_settings'],
          $container->get('logger.factory')
      );
  }

  /**
   * {@inheritdoc}
   *
   * Default formatter configurations:
   * - Youtube_video_play: Play button text
   * - Youtube_poster_quality: Thumbnail quality (hqdefault by default)
   * - Youtube_poster_loading: Image loading mode (lazy by default)
   * - Youtube_nocookie: Use nocookie domain for enhanced privacy
   * - Youtube_autoload: Automatic loading with Intersection Observer
   * - Youtube_short: Show as YouTube Short on mobile devices
   * - Youtube_params: Additional parameters for video URL.
   *
   * @return array
   *   A list of default settings, keyed by the setting name.
   */
  public static function defaultSettings(): array {
    return [
      'youtube_video_play' => \t('Play'),
      'youtube_poster_quality' => 'hqdefault',
      'youtube_poster_loading' => 'lazy',
      'youtube_nocookie' => FALSE,
      'youtube_autoload' => FALSE,
      'youtube_short' => FALSE,
      'youtube_params' => '',
    ] + parent::defaultSettings();
  }

  /**
   * {@inheritdoc}
   *
   *  Generates the formatter configuration form with the following fields:
   *  - Text field to customize the play button
   *  - Thumbnail quality selector
   *  - Image loading mode selector
   *  - Checkbox to enable nocookie mode
   *  - Checkbox to enable automatic loading
   *  - Checkbox to enable YouTube Short mode
   *  - Text field for additional parameters.
   *
   * @return array
   *   The form definition for the formatter settings.
   */
  public function settingsForm(array $form, FormStateInterface $form_state): array {
    $elements = parent::settingsForm($form, $form_state);

    $elements['youtube_video_play'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Video play'),
      '#description' => $this->t('The title of the play button (for translation).'),
      '#size' => 10,
      '#default_value' => $this->getSetting('youtube_video_play'),
    ];

    $elements['youtube_poster_quality'] = [
      '#type' => 'select',
      '#title' => $this->t('Poster quality'),
      '#description' => $this->t('Set thumbnail poster quality (maxresdefault, sddefault, mqdefault, hqdefault).'),
      '#options' => $this->getPosterQualities(),
      '#default_value' => $this->getSetting('youtube_poster_quality'),
    ];

    $elements['youtube_poster_loading'] = [
      '#type' => 'select',
      '#title' => $this->t('Poster loading'),
      '#description' => $this->t('Set img lazy load attr loading for poster image.'),
      '#options' => [
        'eager' => $this->t('eager'),
        'lazy' => $this->t('lazy'),
      ],
      '#default_value' => $this->getSetting('youtube_poster_loading'),
    ];

    $elements['youtube_nocookie'] = [
      '#type' => 'checkbox',
      '#description' => $this->t('Use youtube-nocookie.com as iframe embed uri.'),
      '#title' => $this->t('Nocookie'),
      '#default_value' => $this->getSetting('youtube_nocookie'),
    ];

    $elements['youtube_autoload'] = [
      '#type' => 'checkbox',
      '#description' => $this->t('Use Intersection Observer to load iframe when scrolled into view.'),
      '#title' => $this->t('Autoload'),
      '#default_value' => $this->getSetting('youtube_autoload'),
    ];

    $elements['youtube_short'] = [
      '#type' => 'checkbox',
      '#description' => $this->t('Show 9:16 YouTube Shorts-style interaction on mobile devices.'),
      '#title' => $this->t('Short'),
      '#default_value' => $this->getSetting('youtube_short'),
    ];

    $elements['youtube_params'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Params'),
      '#description' => $this->t('Set YouTube query parameters'),
      '#default_value' => $this->getSetting('youtube_params'),
    ];
    return $elements;
  }

  /**
   * {@inheritdoc}
   *
   * @return array
   *   A short summary of the formatter settings, as an array of strings.
   */
  public function settingsSummary(): array {
    $videoPlay = $this->getSetting('youtube_video_play');
    $posterQuality = $this->getSetting('youtube_poster_quality');
    $posterQualities = $this->getPosterQualities();
    $posterQualityLabel = $posterQualities[$posterQuality] ?? '';
    $posterLoading = $this->getSetting('youtube_poster_loading');
    return [
      $this->t('Video Play: @videoPlay', ['@videoPlay' => $videoPlay]),
      $this->t('Poster Quality: @posterQuality', ['@posterQuality' => $posterQualityLabel]),
      $this->t('Poster Loading: @posterLoading', ['@posterLoading' => $posterLoading]),
    ];
  }

  /**
   * {@inheritdoc}
   */
  public function prepareView(array $entities_items) {
    // No specific preparation needed for this formatter.
  }

  /**
   * {@inheritdoc}
   *
   * Processes each field item to generate the YouTube video rendering.
   *
   * Rendering process:
   * 1. Identifies field type and extracts video ID
   * 2. Validates that the video ID is valid
   * 3. Generates the render array with applied configurations
   * 4. Handles errors and generates error elements when necessary.
   *
   * @return array
   *   A renderable array for $items, as an array of child elements keyed
   *   By consecutive integer starting from 0.
   *
   * @throws \Exception
   *    When the YouTube module is not available or enabled.
   * @throws \InvalidArgumentException
   *    When no valid video ID can be extracted from the field value.
   */
  public function viewElements(FieldItemListInterface $items, $langcode): array {
    $element = [];
    $settings = $this->getSettings();

    foreach ($items as $delta => $item) {
      try {
        if ($item instanceof YouTubeItem) {
          $video_id = $item->video_id;
        }
        else {
          $main_property = $item->getFieldDefinition()
            ->getFieldStorageDefinition()
            ->getMainPropertyName();
          $value = $item->{$main_property};

          if (!\function_exists('youtube_get_video_id')) {
            throw new \Exception('youtube_get_video_id function is not available. Is the YouTube module enabled?');
          }

          $video_id = \youtube_get_video_id($value);
        }

        if (!$video_id) {
          throw new \InvalidArgumentException('No valid video ID extracted.');
        }

        $element[$delta] = [
          '#theme' => 'lite_youtube',
          '#video_id' => $video_id,
          '#params' => $settings['youtube_params'] ?? '',
          '#youtube_poster_quality' => $settings['youtube_poster_quality'] ?? 'sddefault',
          '#youtube_video_play' => $settings['youtube_video_play'] ?? $this->t('Play'),
          '#youtube_poster_loading' => $settings['youtube_poster_loading'] ?? 'lazy',
          '#youtube_nocookie' => $settings['youtube_nocookie'],
          '#youtube_autoload' => $settings['youtube_autoload'],
          '#youtube_short' => $settings['youtube_short'],
          '#entity_title' => $items->getEntity()->label() ?? '',
          '#settings' => $settings,
          '#error' => FALSE,
          '#attached' => [
            'library' => ['lite_youtube/lite_youtube'],
          ],
        ];
      }
      catch (\InvalidArgumentException $e) {
        $this->logger->warning('Lite YouTube formatter skipped item: @message', ['@message' => $e->getMessage()]);
        $element[$delta] = [
          '#theme' => 'lite_youtube',
          '#error' => TRUE,
          '#error_message' => $this->t('Video not available.'),
          '#attached' => [
            'library' => ['lite_youtube/lite_youtube_error'],
          ],
        ];
      }
      catch (\Throwable $e) {
        $this->logger->error('Lite YouTube formatter failed: @message', ['@message' => $e->getMessage()]);
        $element[$delta] = [
          '#theme' => 'lite_youtube',
          '#error' => TRUE,
          '#error_message' => $this->t('An error occurred while loading the video.'),
          '#attached' => [
            'library' => ['lite_youtube/lite_youtube_error'],
          ],
        ];
      }
    }

    return $element;
  }

  /**
   * Get available poster quality options.
   *
   * Returns the available quality options for YouTube thumbnails:
   * - Maxresdefault: Maximum available resolution
   * - Sddefault: Standard definition (640x480)
   * - Mqdefault: Medium quality (320x180)
   * - Hqdefault: High quality (480x360)
   *
   * @return array
   *   An array of poster quality options with their labels.
   */
  protected function getPosterQualities(): array {
    return [
      'maxresdefault' => $this->t('Max'),
      'sddefault' => $this->t('SD'),
      'mqdefault' => $this->t('MQ'),
      'hqdefault' => $this->t('HQ'),
    ];
  }

  /**
   * {@inheritdoc}
   *
   *  Verifies if the formatter can be applied to a specific field:
   *
   *  For entities that are NOT media:
   *  - Only applies to 'youtube' type fields
   *
   *  For media entities:
   *  - Must have a valid media type
   *  - The media type must use an OEmbed source (like YouTube)
   *
   * @return bool
   *   TRUE if the formatter can be used, FALSE otherwise.
   */
  public static function isApplicable(FieldDefinitionInterface $field_definition): bool {
    if ($field_definition->getTargetEntityTypeId() !== 'media') {
      return $field_definition->getType() == 'youtube';
    }

    if (!parent::isApplicable($field_definition)) {
      return FALSE;
    }

    $media_type_id = $field_definition->getTargetBundle();
    if (!$media_type_id) {
      return FALSE;
    }

    $media_type_config = \Drupal::config('media.type.' . $media_type_id);
    $source_plugin_id = $media_type_config->get('source');

    if (!$source_plugin_id) {
      return FALSE;
    }

    $media_source_manager = \Drupal::service('plugin.manager.media.source');

    try {
      $source_plugin_definition = $media_source_manager->getDefinition($source_plugin_id);

      if (!isset($source_plugin_definition['class'])) {
        \Drupal::logger('lite_youtube')->warning('Media source plugin definition missing class key: @plugin_id', [
          '@plugin_id' => $source_plugin_id,
        ]);
        return FALSE;
      }

      $source_plugin_class = $source_plugin_definition['class'];

      return \is_subclass_of($source_plugin_class, OEmbedInterface::class);
    }
    catch (PluginNotFoundException $e) {
      \Drupal::logger('lite_youtube')->warning('Media source plugin not found: @plugin_id for media type @media_type', [
        '@plugin_id' => $source_plugin_id,
        '@media_type' => $media_type_id,
      ]);
      return FALSE;
    }
    catch (\InvalidArgumentException $e) {
      \Drupal::logger('lite_youtube')->warning('Invalid argument in media source check: @message for plugin @plugin_id', [
        '@message' => $e->getMessage(),
        '@plugin_id' => $source_plugin_id,
      ]);
      return FALSE;
    }
  }

}
